import time
import struct
import collections

from scapy.interfaces import show_interfaces
from scapy.layers.inet import IP, UDP
from scapy.sendrecv import sr, sendp, AsyncSniffer, conf
from printer import pass_print, error_print

DEV_IP = None

class UDP_:
    def __init__(self):
        self.receive_cache = collections.deque(maxlen=10000)

        self.iface = '以太网'
        self.dev_ip = '192.168.0.5'
        self.local_ip = None
        self.multicast_ip = '239.255.255.252'
        self.iface_confirmed = False

        self.local_mac = None
        self.dev_mac = 'FF:FF:FF:FF:FF:FF'

        self.local_port = 10001
        self.multicast_port = 10000
        print('Assumption: source port of sensor is ', self.multicast_port, '; destination port of sensor used:', self.local_port)

        self.sniff = None

    def handle_catch_ip_addr(self, packet):
        self.iface_confirmed = True
        self.dev_mac = packet.src
        self.receive_cache.append(bytes(packet)[14:])

    def handle_save_packet(self, packet):
        data = bytes(packet)[42:]
        self.receive_cache.append(data)

    def find_device(self):
        result = False
        self.iface_confirmed = False 

        iface_list = self.get_network_card()
        lens = len(iface_list)
        for i in range(lens):
            self.confirm_iface(iface_list[i])
            if self.iface_confirmed:
                self.local_mac = iface_list[i][1]
                self.local_ip = iface_list[i][2]
                pass_print(f'[NetworkCard]{self.iface}')
                result = True
                break
            else:
                if i == len(iface_list) - 1:
                    result = False
                    error_print('No available Ethernet card was found.')
                    break
        return result

    def confirm_iface(self, iface):
        PK = [0x50, 0x4B]
        cmd_type = struct.unpack('>H', bytes(PK))[0]
        filter_exp = f'ip src {self.dev_ip} and udp and udp port {self.local_port} and ether[42:2] == {cmd_type}'
        payload = self.build_packet(PK, 1, [])
        packet = IP(src=iface[2], dst=self.multicast_ip) / UDP(sport=self.local_port, dport=self.multicast_port)
        packet.add_payload(payload)
        async_sniffer = AsyncSniffer(
            iface=iface[0], prn=self.handle_catch_ip_addr, filter=filter_exp)
        async_sniffer.start()
        time.sleep(0.01)
        sr(packet, iface=iface[0], verbose=0, timeout=1)
        time.sleep(0.5)
        async_sniffer.stop()

        if self.iface_confirmed:
            self.iface = iface[0]
            self.src_mac = iface[1]
            self.local_ip = iface[2]
            
    def get_network_card(self):
        network_card_list = []
        for i in conf.ifaces:
            if conf.ifaces[i].ip == '127.0.0.1' or \
                conf.ifaces[i].mac == '':
                continue
            network_card_list.append(
                (conf.ifaces[i].name, conf.ifaces[i].mac, conf.ifaces[i].ip))
        return network_card_list  

    def sniff_packet(self, type=None):
        if type is None:
            filter_exp = f'ip src {self.dev_ip} and udp and udp port {self.local_port}'
        else:
            cmd_type = struct.unpack('>H', bytes(type))[0]
            filter_exp = f'ip src {self.dev_ip} and udp and udp port {self.local_port} and ether[42:2] == {cmd_type}'
        self.sniff = AsyncSniffer(iface=self.iface, filter=filter_exp, prn=self.handle_save_packet)
        self.sniff.start()
        time.sleep(1)

    def stop_sniff(self):
        if self.sniff is not None:
            self.sniff.stop()
            self.sniff = None

    def fetch_data(self):
            try:
                data = self.receive_cache.popleft()
            except:
                data = None
            return data

    def read_data(self, timeout=10):
        self.sniff_packet()
        start_time = time.time()
        while True:
            if time.time() - start_time < timeout:
                data = self.fetch_data()
                if data is not None:
                    print(data)
            else:
                self.stop_sniff()
                break
    
    def conmunication_check(self, timeout=20):
        self.sniff_packet()        
        start_time = time.time()
        while True:
            if time.time() - start_time < timeout:
                data = self.fetch_data()
                if data is not None:
                    self.stop_sniff()
                    return True
            else:
                self.stop_sniff()
                return False

    def write_read_response(self, type, payload, timeout=1):
        packet = IP(src=self.local_ip, dst=self.multicast_ip) / UDP(sport=self.local_port, dport=self.multicast_port)
        packet.add_payload(payload)
        self.sniff_packet(type)
        sr(packet, iface=self.iface, verbose=0, timeout=1)
        start_time = time.time()
        while True:
            if time.time() - start_time < timeout:
                resp = self.fetch_data()
                if resp is not None:
                    type = resp[:2].decode()
                    cnt = struct.unpack('>I', resp[2:6])[0]
                    len = struct.unpack('>H', resp[6:8])[0]
                    payload = resp[8:8+len]
                    resp = [type, cnt, len, payload]
                    self.stop_sniff()
                    return resp
            else:
                self.stop_sniff()
                return None
                 
    def send_packet(self, payload):
        packet = IP(src=self.local_ip, dst=self.multicast_ip) / UDP(sport=self.local_port, dport=self.multicast_port)
        packet.add_payload(payload)
        sr(packet, iface=self.iface, verbose=0, timeout=0.01)
    
    def build_packet(self, packet_type, cnt, payload):
        '''
        packet_type: *example* [0x50, 0x4B]
        cnt: *example* 1
        payload: *example* b'\x00\x00' or [0x00, 0x00]
        '''
        packet = bytes(packet_type)
        packet += struct.pack('>I', cnt)
        payload_len = struct.pack('>H', len(payload))
        packet += payload_len
        if isinstance(payload, bytes):
            packet += payload
        elif isinstance(payload, list):
            packet += bytes(payload)
        return packet

class Ethernet:
    def __init__(self):
        self.receive_cache = collections.deque(maxlen=1024*16)
        self.src_mac = None
        self.dst_mac = 'FF:FF:FF:FF:FF:FF'
        self.iface = None
        self.local_ip = None
        self.iface_confirmed = False
        self.sniff = None
    
    def handle_catch_packet(self, packet):
        data = bytes(packet)[14:]
        self.receive_cache.append(data)
    
    def handle_catch_mac_addr(self, packet):
        self.iface_confirmed = True
        self.dst_mac = packet.src
        self.receive_cache.append(bytes(packet)[14:])

    def eth_setting_update(self, src, dst, iface):
        self.src_mac = src
        self.dst_mac = dst
        self.iface = iface

    def find_device(self):
        result = False
        self.iface_confirmed = False 

        iface_list = self.get_network_card()
        lens = len(iface_list)
        for i in range(lens):
            self.confirm_iface(iface_list[i])
            if self.iface_confirmed:
                pass_print(f'[NetworkCard]{self.iface}')
                result = True
                break
            else:
                if i == len(iface_list) - 1:
                    result = False
                    error_print('No available Ethernet card was found.')
                    break
        return result

    def confirm_iface(self, iface):
        PING_TYPE = [0x01, 0xcc]
        filter_exp = 'ether dst host ' + \
            iface[1] + ' and ether[16:2] == 0x01cc'
        src_mac = bytes([int(x, 16) for x in iface[1].split(':')])
        dst_mac = self.get_mac_addr(self.dst_mac)
        command_line = self.build_packet(
            PING_TYPE, [], dst_mac, src_mac)
        async_sniffer = AsyncSniffer(
            iface=iface[0], prn=self.handle_catch_mac_addr, filter=filter_exp)
        async_sniffer.start()
        time.sleep(0.01)
        sendp(command_line, iface=iface[0], verbose=0, count=2)
        time.sleep(0.5)
        async_sniffer.stop()

        if self.iface_confirmed:
            self.iface = iface[0]
            self.src_mac = iface[1]
            self.local_ip = iface[2]
            
    def get_network_card(self):
        network_card_list = []
        for i in conf.ifaces:
            if conf.ifaces[i].ip == '127.0.0.1' or \
                conf.ifaces[i].mac == '':
                continue
            network_card_list.append(
                (conf.ifaces[i].name, conf.ifaces[i].mac, conf.ifaces[i].ip))
        return network_card_list

    def sniff_packet(self, type=None):
        if type == None:
            filter_exp = f'ether src host {self.dst_mac} '
        # elif type != None and self.dst_mac == 'FF:FF:FF:FF:FF:FF':
        #     cmd_type = struct.unpack('>H', bytes(type))[0]
        #     filter_exp = f'ether dst host {self.src_mac} and ether[16:2] == {cmd_type}'
        else:
            cmd_type = struct.unpack('>H', bytes(type))[0]
            filter_exp = f'ether src host {self.dst_mac}  and ether[16:2] == {cmd_type}'
        
        if type == [0x01, 0xcc]:
            self.sniff = AsyncSniffer(
                iface=self.iface, prn=self.handle_catch_mac_addr, filter=filter_exp, store=0)
        else:
            self.sniff = AsyncSniffer(
                iface=self.iface, prn=self.handle_catch_packet, filter=filter_exp, store=0)
        self.sniff.start()
        time.sleep(0.1)

    def stop_sniff(self):
        if self.sniff is not None:
            self.sniff.stop()
            self.sniff = None

    def get_mac_addr(self, mac_addr):
        if mac_addr:        
            return bytes([int(x, 16) for x in mac_addr.split(':')])
        return None

    def build_packet(self, packet_type, payload, dst=None, src=None):
        '''
        packet_type: *example* [0x50, 0x4B]
        payload: *example* b'\x00\x00' or [0x00, 0x00]
        '''
        head = b'\x55\x55' 
        packet = bytes(packet_type)
        payload_len = struct.pack('<I', len(payload))
        packet += payload_len
        if isinstance(payload, bytes):
            packet += payload
        elif isinstance(payload, list):
            packet += bytes(payload)
        crc = bytes(self.calc_crc(packet))
        packet += crc
        packet_len = struct.pack('<H', len(packet)+2)
        if src is None:
            src = self.get_mac_addr(self.src_mac)
        if dst is None:
            dst = self.get_mac_addr(self.dst_mac)
        # src = b'\xff\xff\xff\xff\xff\xff'
        packet = dst + src + packet_len + head + packet
        return packet

    def fetch_data(self):
        try:
            data = self.receive_cache.popleft()
        except:
            data = None
        return data

    def write_read_response(self, type, packet, timeout=1):
        self.sniff_packet(type)
        if self.iface is not None:
            sendp(packet, iface=self.iface, verbose=0)
        else:
            sendp(packet, iface='以太网', verbose=0)
        start_time = time.time()
        while True:
            if time.time() - start_time < timeout:
                resp = self.fetch_data()
                if resp is not None:
                    if resp[2:4].hex() == bytes(type).hex():
                        type = resp[2:4].hex()
                        len = struct.unpack('<I', resp[4:8])[0]
                        payload = resp[8:8+len]
                        resp = [type, len, payload]
                        self.stop_sniff()
                        return resp
            else:
                self.stop_sniff()
                return None

    def send_packet(self, payload):
        if payload is None:
            sendp(payload, iface=self.iface, verbose=0, timeout=0.01)

    def calc_crc(self, payload):
        '''
        Calculates 16-bit CRC-CCITT
        '''
        crc = 0x1D0F
        for bytedata in payload:
            crc = crc ^ (bytedata << 8)
            i = 0
            while i < 8:
                if crc & 0x8000:
                    crc = (crc << 1) ^ 0x1021
                else:
                    crc = crc << 1
                i += 1

        crc = crc & 0xffff
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        return [crc_msb, crc_lsb]